import subprocess
from typing import Tuple


def run(cmd:str ,cwd:str=None, env:dict = None)->Tuple[int,str,str]:
    """Run a shell command and return the return code, stdout and stderr"""
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, env=env)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


from typing import List, Dict, Iterable, Generator, Callable,Any
import datetime,random,re,hashlib,hmac


def groupby(l: Iterable, key: str| Callable) -> Dict:
    """group a list by key
    @param l: list of dict
    @param key: key to group by
    @return: dict of key -> list of item
    """
    if not l:
        return {}
    if key and callable(key):
        d = {}
        for item in l:
            k = key(item)
            if k not in d:
                d[k] = []
            d[k].append(item)
        return d
    d = {}
    for item in l:
        k = item[key]
        if k not in d:
            d[k] = []
        d[k].append(item)
    return d

def indexby(l: Iterable, key: str| Callable , keep_first=True) -> Dict:
    """index a list by key
    @param l: list of dict
    @param key: key to index by
    @return: dict of key -> item
    """
    if not l:
        return {}
    if key and callable(key):
        d = {}
        for item in l:
            k = key(item)
            if k in d:
                if keep_first:
                    continue
            d[k] = item
        return d
    d = {}
    for item in l:
        k = item[key]
        if k in d:
            if keep_first:
                continue
        d[k] = item
    return d

def unique(l:List, fn:Callable=None,keep_first=True) -> List:
    """unique list and keep order"""
    if not l:
        return l
    d = {}
    r = []
    if fn:
        for item in l:
            k = fn(item)
            if k in d:
                if keep_first:
                    continue
            d[k] = True
            r.append(item)
        return r
    for item in l:
        if item in d:
            if keep_first:
                continue
        d[item] = True
        r.append(item)
    return r


def partition(l: Iterable,partition_size:int) -> Generator[List, None, None]:
    """Partition a list into partitions of size partition_size"""
    if l is None:
        return
    cur = []
    i = 0
    for v in l:
        cur.append(v)
        i += 1
        if i == partition_size:
            yield cur
            cur = []
            i = 0
    if cur:
        yield cur


def divide(l: List, num_partitions:int) -> List[List]:
    """Partition a list into num_partitions partitions"""
    if not l:
        return l
    partition_size = len(l)//num_partitions
    if partition_size == 0:
        partition_size = 1
    elif len(l) % num_partitions != 0:
        partition_size += 1
    return [l[i*partition_size : (i+1) * partition_size] for i in range (num_partitions)]

def flattern(l:List)->List:
    """flattern a list
    [1,[2,3],[4],[[2]]] -> [1,2,3,4 ,[2]]
    """
    if not l:
        return l
    return [x for i in l for x in (i if isinstance(i, list) or isinstance(i ,set) else [i])]

def flattern_deep(l:List)->List:
    """flattern a list deeply
    [1,[2,3],[4],[[2]]] -> [1,2,3,4 ,2]
    """
    if not l:
        return l
    return [x for i in l for x in (flattern_deep(i) if isinstance(i, list) or isinstance(i ,set) else [i])]

def index_of(l:List, v:Any|Callable)->int:
    """return the index of the first element , -1 not found """
    if callable(v):
        for i, e in enumerate(l):
            if v(e):
                return i
        return -1
    else:
        for i, e in enumerate(l):
            if e == v:
                return i
        return -1

def last_index_of(l:List, v:Any|Callable)->int:
    """return the index of the last element , -1 not found """
    if callable(v):
        for i, e in enumerate(reversed(l)):
            if v(e):
                return len(l) - i - 1
        return -1
    else:
        for i, e in enumerate(reversed(l)):
            if e == v:
                return len(l) - i - 1
        return -1


def now()->int:
    """return current time in seconds"""
    return int(datetime.datetime.now().timestamp())

def pick(d:Dict, keys:List)->Dict:
    """pick keys from a dict"""
    if not d or not keys:
        return d
    return {k: d[k] for k in keys if k in d}

def omit(d:Dict, keys:List)->Dict:
    """omit keys from a dict"""
    if not d or not keys:
        return d
    return {k: d[k] for k in d if k not in keys}

def compact(l:List)->List:
    """remove None, False, 0 from a list"""
    if not l:
        return l
    return [x for x in l if x]

def shuffle(l:List)->List:
    """shuffle a list"""
    l = [v for v in l]
    random.shuffle(l)
    return l

def sample(l:List, n:int)->List:
    """sample n elements from a list"""
    return random.sample(l, n)

def randstr(n:int, az=True,AZ=True, digits = True)->str:
    """generate a random string with length n"""
    s = ''
    if az:
        s += 'abcdefghijklmnopqrstuvwxyz'
    if AZ:
        s += 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    if digits:
        s += '0123456789'
    return ''.join(random.choices(s, k=n))

def rand(lower=0, upper=None)->int:
    """generate a random number between lower and upper"""
    if upper is None:
        upper = lower
        lower = 0
    return random.randint(lower, upper)

def clamp(n:int|float, lower:int|float, upper:int|float)->int|float:
    """clamp a number between lower and upper"""
    if lower > upper:
        lower, upper = upper, lower
    return max(lower, min(n, upper))


def union(*args:List)->List:
    """union multiple lists"""
    return list(set(flattern(args)))

def intersection(*args:List)->List:
    """intersection multiple lists"""
    return list(set.intersection(*map(set, args)))

def difference(a:List, b:List)->List:
    """difference of two lists"""
    return list(set(a) - set(b))

def argsort(l:List, key:Callable=None)->List:
    """sort a list by key and return the index"""
    if key:
        return sorted(range(len(l)), key=lambda i: key(l[i]))
    return sorted(range(len(l)), key=lambda i: l[i])

def trim_left(s:str, x:str=None)->str:
    """remove leading and trailing spaces"""
    if not x:
        return s.lstrip(x)
    if s.startswith(x):
        s =  s[len(x):]
        return trim_left(s, x)
    return s

def trim_right(s:str, x:str=None)->str:
    """remove trailing spaces"""
    if not x:
        return s.rstrip(x)
    if s.endswith(x):
        s =  s[:-len(x)]
        return trim_right(s, x)
    return s

def trim(s:str, x:str=None)->str:
    """remove leading and trailing"""
    return trim_left(trim_right(s, x), x)

def kv(**kwargs)->Dict:
    """create a new dict with non-None values"""
    return {k:v for k,v in kwargs.items() if v is not None}

def re_groups(s:str, pattern:str)->List[str]:
    """
    match a string with a regex pattern and return the group
    @param s: string
    @param pattern: regex pattern
    @param group: group index, None to return all groups
    @return: list of groups or group

    """
    if not s:
        return s
    m = re.match(pattern, s)
    if not m:
        return []
    return m.groups()

def re_find(s:str, pattern:str)->str:
    """
    match a string with a regex pattern and return the first match
    @param s: string
    @param pattern: regex pattern
    @return: first match
    """
    if not s:
        return s
    m = re.match(pattern, s)
    if not m:
        return ''
    return m.group(0)
    
def md5(s:str | bytes)->str:
    """md5 hash of a string"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    return hashlib.md5(s).hexdigest()

def sha1(s:str | bytes)->str:
    """sha1 hash of a string or bytes"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    return hashlib.sha1(s).hexdigest()

def sha256(s:str | bytes)->str:
    """sha256 hash of a string or bytes"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    return hashlib.sha256(s).hexdigest()

def sha512(s:str | bytes)->str:
    """sha512 hash of a string or bytes"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    return hashlib.sha512(s).hexdigest()

def hmac_sha256(s:str | bytes, key:str | bytes)->str:
    """hmac sha256 hash of a string or bytes"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    if isinstance(key, str):
        key = key.encode()
    return hmac.new(key, s, hashlib.sha256).hexdigest()

def hmac_sha512(s:str | bytes, key:str | bytes)->str:
    """hmac sha512 hash of a string or bytes"""
    if not s:
        return ''
    if isinstance(s, str):
        s = s.encode()
    if isinstance(key, str):
        key = key.encode()
    return hmac.new(key, s, hashlib.sha512).hexdigest()


from setuptools import setup, find_packages

# pip install -m venv venv
# source venv/bin/activate
# pip install pytest
# venv/bin/pytest

# pip install build twine
# python -m build
# python -m twine upload --verbose dist/happydata-0.0.12*
# pip install happydata

setup(
    name='happydata',
    version='0.0.12',
    author="Daqian",
    packages=find_packages(exclude=['tests', 'tests.*','venv']),
    zip_safe=False,
    description='happy data loading,writing, transforming without third party library',
    long_description='happy data loading,writing, transforming without third party library',
    license='MIT license',
    keywords=['data', 'data loading', 'data writing', 'data transforming','json','jsonl','gz','lodash'],
    platforms='Independant',
    url='https://github.com/daqiancode/happydata',
)


import json,os,gzip,io,tarfile,zipfile
from typing import List,Dict,Iterable,Generator,Optional,IO,Tuple,BinaryIO,Any
from pathlib import Path

def write_jsonl(l: Iterable, to:str):
    """write a list of dict to a jsonl file, support both .gz and plain text"""
    if str(to).endswith('.gz'):
        write_jsonl_gz(l, to)
        return
    with open(to, 'w') as f:
        for item in l:
            f.write(json.dumps(item, ensure_ascii=False, default=vars) + '\n')

def read_jsonl(from_file:str |Path| IO)->Generator[Any,str,None]:
    """read a jsonl file as a json object generator, support both .gz and plain text"""
    if isinstance(from_file, str) or isinstance(from_file, Path):
        if str(from_file).endswith('.gz'):
            with gzip.open(from_file, 'rt') as f:
                for line in f:
                    yield json.loads(line)
        else:
            with open(from_file, 'r') as f:
                for line in f:
                    yield json.loads(line)
    else:
        for line in from_file:
            yield json.loads(line)

def write_jsonl_gz(l: Iterable, to:str, add_end=True):
    """write a list of dict to a jsonl file"""
    with gzip.open(to, 'wt') as f:
        for item in l:
            if add_end:
                f.write(json.dumps(item, ensure_ascii=False, default=vars) + '\n')
            else:
                f.write(json.dumps(item, ensure_ascii=False, default=vars))

def read_jsonl_gz(from_file:str|Path| IO)->Generator[Any,str,None]:
    """read a jsonl file as a json object generator"""
    if isinstance(from_file, str) or isinstance(from_file, Path):
        with gzip.open(from_file, 'rt') as f:
            for line in f:
                yield json.loads(line)
    else:
        for line in from_file:
            yield json.loads(line)

def load_jsonl(from_file:str)->List[Any]:
    """load a jsonl file to memory as a list of json object, support both .gz and plain text"""
    return list(read_jsonl(from_file))

def load_jsonl_gz(from_file:str)->List[Dict]:
    """load a jsonl file to memory as a list of json object"""
    return list(read_jsonl_gz(from_file))

def load_lines(file:str, remove_end=True)->List[str]:
    """load a text file to memory as a list of string, support both .gz and plain text"""
    return list(read_lines(file , remove_end))

def read_lines(file:str|Path,remove_end=True)->Generator[str,str,None]:
    """read a text file as a line generator, remove \n and \r by default. support both .gz and plain text"""
    if str(file).endswith('.gz'):
        with gzip.open(file, 'rt') as f:
            for line in f:
                if remove_end:
                    yield line.removesuffix('\n').removesuffix('\r')
                else:
                    yield line
    else:
        with open(file, 'r') as f:
            for line in f:
                if remove_end:
                    yield line.removesuffix('\n').removesuffix('\r')
                else:
                    yield line

def read_lines_gz(file:str,remove_end=True)->Generator[str,str,None]:
    """read a text file as a line generator, remove \n and \r by default"""
    with gzip.open(file, 'rt') as f:
        for line in f:
            if remove_end:
                yield line.removesuffix('\n').removesuffix('\r')
            else:
                yield line

def write_lines(lines:Iterable[str],to:str|Path ,add_end=True):
    """write a list of string to a text file, support both .gz and plain text"""
    if str(to).endswith('.gz'):
        write_lines_gz(lines, to, add_end)
        return
    with open(to, 'w') as f:
        for line in lines:
            if add_end:
                f.write(str(line)+'\n')
            else:
                f.write(str(line))
def write_lines_gz(lines:Iterable[str],to:str|Path ,add_end=True):
    with gzip.open(to, 'wt') as f:
        for line in lines:
            if add_end:
                f.write(str(line)+'\n')
            else:
                f.write(str(line))

def read_tar_gz(file:str|Path) -> Generator[Tuple[tarfile.TarInfo,IO],str,None]:
    """read a tar.gz file as a file generator"""
    with tarfile.open(file, 'r:gz') as tar:
        for member in tar.getmembers():
            with tar.extractfile(member) as f:
                yield member , f

def write_tar_gz(file:str|Path,files:Dict[str,BinaryIO]):
    """write a dict of file to a tar.gz file"""
    with tarfile.open(file, 'w:gz') as tar:
        for name, f in files.items():
            info = tarfile.TarInfo(name)
            buf = f.read()
            info.size = len(buf)
            tar.addfile(info, io.BytesIO(buf))

def read_zip(file:str|Path) -> Generator[Tuple[str,IO],str,None]:
    """read a zip file as a file generator"""
    with zipfile.ZipFile(file, 'r') as z:
        for member in z.namelist():
            with z.open(member) as f:
                yield member , f

def write_zip(file:str|Path,files:Dict[str,BinaryIO]):
    """write a dict of file to a zip file"""
    with zipfile.ZipFile(file, 'w') as z:
        for name, f in files.items():
            z.writestr(name, f.read())

def load_text(file:str|Path)->str:
    """file to text"""
    if str(file).endswith('.gz'):
        with gzip.open(file, 'rt') as f:
            return f.read()
    with open(file, 'r') as f:
        return f.read()

def write_text(s:str,to:str|Path):
    """text to file"""
    if str(to).endswith('.gz'):
        with gzip.open(to, 'wt') as f:
            f.write(s)
        return
    with open(to, 'w') as f:
        f.write(s)

def load_json(file:str|Path)->Dict:
    """file to json"""
    if str(file).endswith('.gz'):
        with gzip.open(file, 'rt') as f:
            return json.load(f)
    with open(file, 'r') as f:
        return json.load(f)

def write_json(d:Dict,to:str|Path):
    """json to file"""
    if str(to).endswith('.gz'):
        with gzip.open(to, 'wt') as f:
            json.dump(d, f,ensure_ascii=False, default=vars)
        return
    with open(to, 'w') as f:
        json.dump(d, f,ensure_ascii=False, default=vars)

def match_suffix(file:str|Path,suffix:str|List)->bool:
    """check if a file has a suffix"""
    if isinstance(suffix, str):
        return str(file).lower().endswith(suffix.lower())
    if isinstance(suffix, list):
        for s in suffix:
            if str(file).lower().endswith(s.lower()):
                return True
        return False

def read_dir(d:str|Path, recursive:bool=False,suffix:str|List=None)->Generator[str,str,None]:
    """list all files in a directory in absolute path"""
    if not recursive:
        for f in os.listdir(d):
            if suffix:
                if os.path.isfile(os.path.join(d,f)) and match_suffix(f, suffix):
                    yield os.path.abspath(os.path.join(d,f))
            else:
                if os.path.isfile(os.path.join(d,f)):
                    yield os.path.abspath(os.path.join(d,f))
    else:
        for root, dirs, files in os.walk(d):
            for f in files:
                if suffix:
                    if match_suffix(f, suffix):
                        yield os.path.abspath(os.path.join(root,f))
                else:
                    yield os.path.abspath(os.path.join(root,f))

def load_dir(d:str|Path, recursive:bool=False,suffix:str=None)->List[str]:
    """list all files in a directory in absolute path"""
    return list(read_dir(d, recursive, suffix))


def read_tar(file:str|Path) -> Generator[Tuple[tarfile.TarInfo,IO],str,None]:
    """read a tar file as a file generator"""
    with tarfile.open(file, 'r') as tar:
        for member in tar.getmembers():
            with tar.extractfile(member) as f:
                yield member , f

def write_tar(file:str|Path,files:Dict[str,BinaryIO]):
    """write a dict of file to a tar file"""
    with tarfile.open(file, 'w') as tar:
        for name, f in files.items():
            info = tarfile.TarInfo(name)
            buf = f.read()
            info.size = len(buf)
            tar.addfile(info, io.BytesIO(buf))